import os
from glob import glob
import numpy as np
import pandas as pd
import shutil
import json
from datetime import datetime

from .BaseStorage import BaseStorage
from .decorators import with_rio, with_geopandas, with_xarray, with_rioxarray


class LocalStorage(BaseStorage):
    def __init__(self, path="data"):
        super().__init__()
        self.path = path
        if not os.path.exists(self.path):
            os.makedirs(self.path)
            print(f"'{self.path}' created")
        else:
            print(f"'{self.path}' folder ready to use")

    def exists(self, name):
        return os.path.exists(self.get_path(name))

    def get_path(self, name):
        return os.path.join(self.path, name)

    def delete(self, name):
        if os.path.isdir(self.get_path(name)):
            shutil.rmtree(self.get_path(name))
        else:
            os.remove(self.get_path(name))

    def create_from_path(self, data, name):
        dst_path = self.get_path(name)
        shutil.move(data, dst_path)  # porque se hace un move en vez de un copy?
        return dst_path

    def create_from_dict(self, data, name):
        if name.endswith(".json") or name.endswith(".geojson"):
            dst_path = self.get_path(name)
            self.create_dir(dst_path)
            with open(dst_path, "w") as f:
                json.dump(data, f)
            return dst_path
        else:
            raise TypeError("Not a valid dict type extension")

    def create_from_string(self, data, name):
        dst_path = self.get_path(name)
        self.create_dir(dst_path)
        with open(dst_path, "w") as f:
            f.write(data)
        return dst_path

    def create_from_dataframe(self, data, name):
        if name.endswith(".csv"):
            dst_path = self.get_path(name)
            self.create_dir(dst_path)
            data.to_csv(dst_path)
            return dst_path
        elif name.endswith(".json"):
            dst_path = self.get_path(name)
            self.create_dir(dst_path)
            data.to_json(dst_path)
            return dst_path
        elif name.endswith(".geojson"):
            dst_path = self.get_path(name)
            self.create_dir(dst_path)
            data.to_file(dst_path)
            return dst_path

    def create_from_image(self, data, name):
        dst_path = self.get_path(name)
        self.create_dir(dst_path)
        data.save(dst_path)
        return dst_path

    @with_rio
    def create_from_rasterio(self, rio, x, name, ds, window=None):
        dst_path = self.get_path(name)
        self.create_dir(dst_path)
        kwargs = ds.meta.copy()
        transform = ds.transform if window is None else ds.window_transform(window)
        kwargs.update(
            driver="GTiff",
            count=1 if x.ndim < 3 else x.shape[0],
            height=x.shape[0] if x.ndim < 3 else x.shape[1],
            width=x.shape[1] if x.ndim < 3 else x.shape[2],
            dtype=np.uint8 if x.dtype == "bool" else x.dtype,
            crs=ds.crs,
            transform=transform,
            # nbits=1 if x.dtype == 'bool' else
        )
        with rio.open(dst_path, "w", **kwargs) as dst:
            bands = 1 if x.ndim < 3 else [i + 1 for i in range(x.shape[0])]
            dst.write(x, bands)
        return dst_path

    def create_from_array(self, data, name):
        dst_path = self.get_path(name)
        self.create_dir(dst_path)
        np.save(dst_path, data)
        return dst_path

    def create_from_csv(self, data, name):
        dst_path = self.get_path(name)
        self.create_dir(dst_path)
        data.to_csv(dst_path)
        return dst_path

    def create_from_json(self, data, name):
        dst_path = self.get_path(name)
        self.create_dir(dst_path)
        data.to_json(dst_path)
        return dst_path

    def create_from_parquet(self, data, name):
        dst_path = self.get_path(name)
        self.create_dir(dst_path)
        data.to_parquet(dst_path)
        return dst_path

    def create_from_zarr(self, data, name):
        dst_path = self.get_path(name)
        self.create_dir(dst_path)
        data.to_zarr(dst_path)
        return dst_path

    @with_rioxarray
    def create_from_xarray(self, rxr, data, name):
        dst_path = self.get_path(name)
        self.create_dir(dst_path)
        data.rio.to_raster(dst_path)
        return dst_path

    def list(self, pattern="*", recursive=True):
        paths = glob(os.path.join(self.path, pattern))
        # strip base path
        return [p.replace(self.path + "/", "") for p in paths]

    def read_from_array(self, name, path=None):
        if path is None:
            path = self.get_path(name)
        return np.load(path)

    @with_rio
    def read_from_rasterio(self, rio, name):
        return rio.open(self.get_path(name))

    def read_from_csv(self, name):
        return pd.read_csv(self.get_path(name), index_col=0)

    def read_from_json(self, name):
        return pd.read_json(self.get_path(name))

    @with_geopandas
    def read_from_geojson(self, gpd, name):
        return gpd.read_file(self.get_path(name))

    @with_geopandas
    def read_from_parquet(self, gpd, name):
        return gpd.read_parquet(self.get_path(name))

    @with_xarray
    def read_from_zarr(self, xr, name):
        return xr.open_zarr(self.get_path(name))



# from .download import download_satellite_image, download_dem, download_cloud_mask
# from .explore import explore_satellite_images

from .stac import (
    AWSS2L2ADownloader,
    PCS1GRDDownloader,
    PCS1RTCDownloader,
    AWSDEM30Downloader,
    AWSDEM90Downloader,
    PCESAWCDownloader,
    PCLandsat8Downloader,
    PCModisBurnedDownloader,
    PCModisSnow8Downloader,
    PCModisSnowDayDownloader,
    PCAlosPalsarAnnualDownloader,
)

DOWNLOADERS = {
    "sentinel-2-l2a": AWSS2L2ADownloader,
    "sentinel-1-grd": PCS1GRDDownloader,
    "sentinel-1-rtc": PCS1RTCDownloader,
    "cop-dem-glo-30": AWSDEM30Downloader,
    "cop-dem-glo-90": AWSDEM90Downloader,
    "esa-worldcover": PCESAWCDownloader,
    "landsat-8-c2-l2": PCLandsat8Downloader,
    "modis-burned-areas": PCModisBurnedDownloader,
    "modis-snow-cover-8": PCModisSnow8Downloader,
    "modis-snow-cover-daily": PCModisSnowDayDownloader,
    "alos-palsar-mosaic": PCAlosPalsarAnnualDownloader,
}
AVAILABLE_COLLECTIONS = list(DOWNLOADERS.keys())

from .download_stac import (
    download_satellite_imagery,
    load_satellite_imagery,
    explore_satellite_imagery,
)



"""
Decorators for the STAC module.
"""


def with_rioxarray(func):
    def wrapper(self, *args, **kwargs):
        try:
            import rioxarray as rxr

            return func(self, *args, **kwargs)
        except ImportError:
            raise ImportError(
                "The rioxarray and dask diagnostics packages are required. Please install them with 'pip install rioxarray dask[diagnostics]' and try again."
            )

    return wrapper



import os
import json
import yaml
from ..models import Config
from rich import print
from typing import List
from pathlib import Path


def validate_folder(dir, folder, typer):
    if not os.path.exists(dir / folder):
        raise typer.BadParameter(f"No {folder} directory found in '{dir}'.")


def validate_item(dir, folder, item, name, typer, file="main.py", check_file=True):
    # check name
    if not item.name:
        raise typer.BadParameter(f"{name} '{item.name}' is missing 'name' attribute.")
    # check folder has folder with item name
    if not os.path.exists(dir / folder / item.name):
        raise typer.BadParameter(
            f"{name} '{item.name}' cannot be found in {dir}/{folder}."
        )
    # check folder has file
    if not file in os.listdir(dir / folder / item.name):
        if check_file:
            raise typer.BadParameter(f"{name} '{item.name}' is missing file 'main.py'.")
        # print(
        #     f"Warning: file 'main.py' not found in {dir}/{folder}/{item.name}, skipping file check."
        # )
    # TODO: check optionals: reqs, env...


def validate_storage(storage, typer, cloud):
    if storage.type == "local" and cloud:
        raise typer.BadParameter(
            f"Local storage not allowed in cloud deployment, please use S3 storage."
        )
    # if storage.type == "s3" and not storage.credentials:
    #     typer.echo(
    #         f"S3 storage credentials not provided, an S3 bucket will be created if it does not already exist."
    #     )


def load_and_validate_config(dir, typer, verbose=False, cloud=False):
    dir = Path(dir).resolve()
    # check dir exists
    if not dir.exists():
        raise typer.BadParameter(f"Directory '{dir}' does not exist.")
    # check dir is a spai project
    if not "spai.config.yaml" in os.listdir(dir):
        raise typer.BadParameter(
            f"Directory '{dir}' is not a spai project. No spai.config.yaml file found."
        )
    # load config
    config = {}
    with open(dir / "spai.config.yaml", "r") as f:
        config = yaml.safe_load(f)
    if not config:
        raise typer.BadParameter(f"spai.config.yaml file is empty.")
    config.update(dir=dir)
    if verbose:
        print(config)
    config = Config(**config)
    # TODO: check if project name is already taken in cloud, locally is not a problem
    config.project = dir.name if not config.project else config.project
    # check storage
    if config.storage:
        names = []
        for storage in config.storage:
            if storage.name in names:
                raise typer.BadParameter(
                    f"Found multiple storages with name '{storage.name}', please use unique names for your storage."
                )
            validate_storage(storage, typer, cloud)
            names.append(storage.name)
    # check scripts
    if config.scripts:
        # check project has scripts folder
        validate_folder(dir, "scripts", typer)
        for script in config.scripts:
            validate_item(dir, "scripts", script, "script", typer)
    # check apis
    if config.apis:
        # check project has apis folder
        validate_folder(dir, "apis", typer)
        for api in config.apis:
            validate_item(dir, "apis", api, "api", typer)
    # check uis
    if config.uis:
        # check project has uis folder
        validate_folder(dir, "uis", typer)
        for ui in config.uis:
            validate_item(dir, "uis", ui, "ui", typer, check_file=False)
    # check notebooks
    if config.notebooks:
        # check project has notebooks folder
        validate_folder(dir, "notebooks", typer)
        for notebook in config.notebooks:
            validate_item(dir, "notebooks", notebook, "notebook", typer, "main.ipynb")
    return config



"""
AWS module for downloading Sentinel-2 Level-2A data from Earth Search.
"""

from .AWSS2L2ADownloader import AWSS2L2ADownloader
from .AWSDEM30Downloader import AWSDEM30Downloader
from .AWSDEM90Downloader import AWSDEM90Downloader



from spai.storage import Storage
from spai.data.satellite import explore_satellite_imagery, download_satellite_imagery

# explore available images
print("Looking for images in the last month")
bbox = (2.0052191, 41.38238988, 2.17002873, 41.48530451)  # Collserola
images = explore_satellite_imagery(bbox, cloud_cover=10)
if len(images) == 0:
    raise ValueError("No images found")

# download images and save locally
storage = Storage()
sensor = "sentinel-2-l2a"
existing_images = storage["data"].list(f"{sensor}*.tif")
dates = [image.split("_")[1].split(".")[0] for image in existing_images]
print("Found", len(images), f"image{'s' if len(images) > 1 else ''}")
new_images = []
for image in images:
    # check if image is already downloaded
    date = image["datetime"].split("T")[0]
    if date in dates or date in new_images:
        print("Image already downloaded:", date)
        continue
    new_images.append(date)
    print("Downloading new image:", date)
    path = download_satellite_imagery(storage["data"], bbox, date, sensor)
    print("Image saved at", path)



from ..repos import APIRepo


def get_services(user, project_name, format=False):
    repo = APIRepo()
    data, error = repo.retrieve_project_by_name(user, project_name)
    services = []
    if data and "services" in data:
        if not format:
            return data["services"]
        for service in data["services"]:
            data, error = repo.retrieve_service(user, service["id"])
            if error:
                print("Something went wrong.\n" + error)
            if data["type"] == "s3":
                services.append(
                    {
                        "type": data["type"],
                        "url": data["url"],
                        "bucket": data["bucket"],
                        "region": data["region"],
                    }
                )
                continue
            if data["type"] == "script":
                services.append(
                    {
                        "type": data["type"],
                        "name": data["name"],
                    }
                )
                continue
            if (
                data["type"] == "api"
                or data["type"] == "ui"
                or data["type"] == "notebook"
            ):
                services.append(
                    {
                        "type": data["type"],
                        "name": data["name"],
                        "url": data["url"],
                    }
                )
                continue
            raise Exception("Invalid service type")
        return services
    raise Exception("Something went wrong.\n" + error)



"""
PCLandsat8Downloader class to download and load Landsat 8 data from a STAC API of the Planetary Computer.

https://planetarycomputer.microsoft.com/dataset/landsat-c2-l2
"""

from .PCDownloader import PCDownloader

from typing import Optional, List


LANDSAT_8_BANDS = ["coastal", "blue", "green", "red", "nir08", "swir16", "swir22"]


class PCLandsat8Downloader(PCDownloader):
    def __init__(
        self,
        aoi,
        datetime,
        bands: Optional[List[str]] = LANDSAT_8_BANDS,
        resolution: Optional[float] = 0.0000275,
        crs: Optional[str] = "EPSG:4326",
        cloud_cover: Optional[float] = 100,
    ):
        # TODO add filters
        query = {
            "eo:cloud_cover": {"lt": cloud_cover},
        }
        super().__init__(aoi, datetime, query)
        self.collection = "landsat-8-c2-l2"
        self.bands = bands
        self.resolution = resolution
        self.crs = crs



# Updated read_raster function to handle invalid extensions
def read_raster(image_name, storage, bands=None):
    ds = storage.read(image_name)
    if ds is None:
        raise TypeError("Not a valid raster file")
    if not bands:
        raster = ds.read()
    else:
        raster = ds.read(bands)
    return ds, raster



"""
Explore, load and download satellite imagery data
"""

import numpy as np
from typing import List, Union, Optional, Any
from datetime import datetime

from . import DOWNLOADERS, AVAILABLE_COLLECTIONS
from .utils import (
    create_aoi_geodataframe,
    get_last_month,
    add_item_extra_properties,
)

from ...storage import Storage


def load_satellite_imagery(
    aoi: Any,
    date: Optional[List[Union[str, datetime]]] = None,
    collection: str = "sentinel-2-l2a",
    clip: Optional[bool] = False,
    crs: Optional[str] = "epsg:4326",
    **kwargs,
):
    """
    Load satellite imagery data from a given area of interest (aoi) and date in memory.

    Parameters
    ----------
    aoi : Any
        Area of interest. It can be a GeoDataFrame, a list of coordinates, a bounding box, etc.
    date : Optional[List[Union[str, datetime]]], optional
        Date of the image, by default None. If None, the last available image will be loaded.
    collection : str, optional
        Satellite collection to download, by default "sentinel-2-l2a"
    clip : Optional[bool], optional
        Clip the data to the area of interest, by default False
    crs : Optional[str], optional
        Coordinate Reference System, by default "epsg:4326"
    kwargs : dict
        Extra parameters to pass to the downloader, such as bands, cloud_cover, vegetation_percentage an so on.   TODO

    Returns
    -------
    xarray.Dataset
        Satellite imagery data in memory.
    """
    gdf = create_aoi_geodataframe(aoi, crs)
    bbox = gdf.total_bounds
    if collection not in AVAILABLE_COLLECTIONS:
        raise ValueError(
            f"Collection {collection} not available. Available collections are: {AVAILABLE_COLLECTIONS}"
        )

    if not date and "dem" not in collection:
        # If no date given, search for the last available image
        # until a valid image is found
        last_available_image_date = None
        while not last_available_image_date:
            date = get_last_month(date)
            search = explore_satellite_imagery(gdf, date, collection, crs, **kwargs)
            if search:
                last_available_image_date = search[-1]["datetime"]
                date = last_available_image_date

    downloader = DOWNLOADERS[collection](bbox, date, **kwargs)
    data = downloader.load_stac()

    if clip:
        data = downloader.clip_data(data, gdf)

    return data


def download_satellite_imagery(
    storage: Storage,
    aoi: Any,
    date: Optional[List[Union[str, datetime]]] = None,
    collection: str = "sentinel-2-l2a",
    name: Optional[str] = None,
    clip: Optional[bool] = False,
    crs: Optional[str] = "epsg:4326",
    **kwargs,
) -> List[str]:
    """
    Download satellite imagery data from a given area of interest (aoi) and date to a given storage.

    Parameters
    ----------
    storage : Storage
        Storage object to save the data.
    aoi : Any
        Area of interest. It can be a GeoDataFrame, a list of coordinates, a bounding box, etc.
    date : Optional[List[Union[str, datetime]]], optional
        Date of the image, by default None. If None, the last available image will be loaded.
    collection : str, optional
        Satellite collection to download, by default "sentinel-2-l2a"
    clip : Optional[bool], optional
        Clip the data to the area of interest, by default False
    crs : Optional[str], optional
        Coordinate Reference System, by default "epsg:4326"
    kwargs : dict
        Extra parameters to pass to the downloader, such as bands, cloud_cover, vegetation_percentage an so on.   TODO

    Returns
    -------
    None
    """
    data = load_satellite_imagery(aoi, date, collection, clip, crs, **kwargs)
    if not data:
        return None
    data = data.compute()
    paths = []
    for date in data.time.values:
        date = np.datetime_as_string(date, unit="D")
        path = storage.create(
            data.sel(time=date).squeeze(),
            name=name if name else f"{collection}_{date}.tif",
        )
        paths.append(path)
    return paths if len(paths) > 1 else paths[0]


def explore_satellite_imagery(
    aoi: Any,
    date: Optional[List[Union[str, datetime]]] = None,
    collection: str = "sentinel-2-l2a",
    crs: Optional[str] = "epsg:4326",
    **kwargs,
) -> dict:
    """
    Explore satellite imagery data from a given area of interest (aoi) and date.

    Parameters
    ----------
    aoi : Any
        Area of interest. It can be a GeoDataFrame, a list of coordinates, a bounding box, etc.
    date : Optional[List[Union[str, datetime]]], optional
        Date of the image, by default None. If None, the available images of the last month will be loaded.
    collection : str, optional
        Satellite collection to download, by default "sentinel-2-l2a"
    crs : Optional[str], optional
        Coordinate Reference System, by default "epsg:4326"
    kwargs : dict
        Extra parameters to pass to the downloader, such as bands, cloud_cover, vegetation_percentage an so on.   TODO

    Returns
    -------
    dict
        Information about the available images. If no image is found, it returns None. If any extra parameter is given, it will be added to the output.
    """
    gdf = create_aoi_geodataframe(aoi, crs)
    bbox = gdf.total_bounds
    if collection not in AVAILABLE_COLLECTIONS:
        raise ValueError(
            f"Collection {collection} not available. Available collections are: {AVAILABLE_COLLECTIONS}"
        )

    if not date:
        date = get_last_month()

    downloader = DOWNLOADERS[collection](bbox, date, **kwargs)
    search = downloader.search_stac()

    if not search:
        return None

    search_list = []
    for item in search:
        item_dict = {
            "id": item.id,
            "datetime": item.datetime.strftime("%Y-%m-%dT%H:%M:%SZ"),
        }
        item_dict["thumbnail"] = (
            item.assets["thumbnail"].href if "thumbnail" in item.assets else None
        )
        # Add extra parameters from the query, such as cloud cover, etc.
        item_dict = add_item_extra_properties(item, item_dict, downloader.query, kwargs)
        search_list.append(item_dict)
    # Sort by datetime
    search_list_sorted = sorted(search_list, key=lambda x: x["datetime"])

    return search_list_sorted



"""
Planetary Computer (PC) STAC API for Planetary Computer data.
"""

from .PCS1GRDDownloader import PCS1GRDDownloader
from .PCS1RTCDownloader import PCS1RTCDownloader
from .PCESAWCDownloader import PCESAWCDownloader
from .PCLandsat8Downloader import PCLandsat8Downloader
from .PCModisBurnedDownloader import PCModisBurnedDownloader
from .PCModisSnow8Downloader import PCModisSnow8Downloader
from .PCModisSnowDayDownloader import PCModisSnowDayDownloader
from .PCAlosPalsarAnnualDownloader import PCAlosPalsarAnnualDownloader



import matplotlib
import numpy as np
from warnings import warn


def colorize_raster(raster, colors, colorize_zero=False):
    """
    Colorize a raster image based on unique values in the raster.

    Parameters
    ----------
    raster : numpy.ndarray
        A 2-dimensional or 3-dimensional array representing a raster image.
    colors : dict or list
        A dictionary mapping unique values in the raster to colors in the format {value: color} or a list of colors.
        If a list is provided, the colors will be assigned to unique values in the raster in ascending order.
    colorize_zero : bool, optional
        Whether to colorize 0 values in the raster. Default is False.

    Returns
    -------
    numpy.ndarray
        A 4-dimensional array representing the colorized raster image with shape (4, height, width).
    """
    # If raster has 3 dimensions, select the first band
    if raster.ndim == 3:
        raster = raster[0, :, :]
    elif raster.ndim != 2:
        raise ValueError("Input raster must be a 2-dimensional or 3-dimensional array")

    # Get unique values in the raster excluding 0 if colorize_zero is False
    if not colorize_zero:
        unique_values = np.unique(raster)
        unique_values = unique_values[unique_values > 0]
    else:
        unique_values = np.unique(raster)

    if isinstance(colors, dict):
        color_mapping = colors
        # Ensure all unique values have a color in the mapping
        for value in unique_values:
            if value not in color_mapping:
                warn(
                    f"No color provided for raster value {value} in the color mapping."
                )
    else:
        if len(unique_values) != len(colors):
            warn(
                f"The number of unique values in the raster ({len(unique_values)}) does not match the number of colors provided ({len(colors)})."
            )
        color_mapping = {value: color for value, color in zip(unique_values, colors)}

    # Create an array for colors with 4 bands (RGB + Alpha)
    colored_image = np.zeros((raster.shape[0], raster.shape[1], 4), dtype=np.uint8)

    # Map colors to the image
    for value, color in color_mapping.items():
        rgb_color = np.array(matplotlib.colors.to_rgb(color)) * 255
        colored_image[raster == value, :3] = rgb_color

    # Initialize the alpha channel to 255 (fully opaque)
    colored_image[:, :, 3] = 255

    if not colorize_zero:
        # Set alpha to 0 where RGB values are [0, 0, 0]
        mask = np.all(colored_image[:, :, :3] == 0, axis=-1)
        colored_image[mask, 3] = 0

    # transpose
    colored_image = np.transpose(colored_image, (2, 0, 1))

    return colored_image



"""
Module to download satellite imagery data from STAC API.
"""

from .aws import AWSS2L2ADownloader, AWSDEM30Downloader, AWSDEM90Downloader
from .pc import (
    PCS1GRDDownloader,
    PCS1RTCDownloader,
    PCESAWCDownloader,
    PCLandsat8Downloader,
    PCModisBurnedDownloader,
    PCModisSnow8Downloader,
    PCModisSnowDayDownloader,
    PCAlosPalsarAnnualDownloader,
)



"""
PCModisSnow8Downloader class to download and load MODIS Snow Cover 8-day data from a STAC API of the Planetary Computer.

https://planetarycomputer.microsoft.com/dataset/modis-10A2-061
"""

from .PCDownloader import PCDownloader

from typing import Optional, List


MODIS_SNOW_8_BANDS = ["Maximum_Snow_Extent", "Eight_Day_Snow_Cover"]


class PCModisSnow8Downloader(PCDownloader):
    def __init__(
        self,
        aoi,
        datetime,
        bands: Optional[List[str]] = MODIS_SNOW_8_BANDS,
        resolution: Optional[float] = 0.00587,
        crs: Optional[str] = "EPSG:4326",
    ):
        # TODO add filters
        query = {}
        super().__init__(aoi, datetime, query)
        self.collection = "modis-10A2-061"
        self.bands = bands
        self.resolution = resolution
        self.crs = crs



from ..repos import APIRepo
from ..models.StorageConfig import S3StorageCredentials
from ..auth import auth


def create_or_retrieve_s3_bucket(project_name, storage_name):
    user = auth()
    repo = APIRepo()
    data, error = repo.create_or_retrieve_s3_bucket(user, project_name, storage_name)
    if data:
        return S3StorageCredentials(**data)
    raise Exception("Something went wrong.\n" + error)



from spai.storage import Storage
import numpy as np
import pandas as pd

storage = Storage()["data"]


def ndvi(image):
    ds = storage.read(image)
    red = ds.read(4).astype(np.float32)
    nir = ds.read(8)
    ndvi = (nir - red) / (nir + red + 1e-8)
    date = image.split("_")[1].split(".")[0]
    return storage.create(ndvi, f"NDVI_{date}.tif", ds=ds)


def categorize(data, cats=[-100, 0.1, 0.3, 0.5, 100]):
    data0 = data.copy()
    for i, cat in enumerate(cats[:-1]):
        data0[(data > cat) & (data <= cats[i + 1])] = i
    data0 = data0.astype(np.uint8)
    return data0


# retrieve all images
images = storage.list("sentinel-2-l2a*")
print("Found the following images:", images)

# compute ndvis

for image in images:
    ndvi(image)
images = storage.list("NDVI*")
print("NDVI images saved at", images)

# categorize ndvis

for image in images:
    ds = storage.read(image)
    data = ds.read(1)
    cats = categorize(data)
    date = image.split("_")[1].split(".")[0]
    storage.create(cats, f"cat_NDVI_{date}.tif", ds=ds)

images = storage.list("cat_NDV*")

# compute statistics

stats = []
mppx = 10 * 10  # meters per pixel
hasppx = mppx / 10000  # hectares per square meter
for image in images:
    ds = storage.read(image)
    data = ds.read(1)
    counts = np.bincount(data.flatten())
    has = counts * hasppx
    stats.append(has)

dates = [image.split("_")[2].split(".")[0] for image in images]
df = pd.DataFrame(stats, columns=["bare ground", "low", "medium", "high"], index=dates)
df.index = pd.to_datetime(df.index)
df = df.sort_index()

storage.create(df, "stats.csv")



import typer
from pathlib import Path
from typing import List
import json

# Add the cli directory to the Python path
# spai_cli_dir = os.path.dirname(os.path.realpath(__file__))
# sys.path.append(os.path.join(spai_cli_dir))

from .commands import auth as _auth
from .commands import list
from .repos import APIRepo
from .project import (
    init_project,
    install_requirements,
    run_local,
    deploy_folder,
    get_logs,
    download_template,
    deploy_template,
    get_services,
    stop_service,
    stop_service_by_name,
    delete_project,
    get_projects,
)
from .auth import auth
from .config import parse_vars
from . import __version__

app = typer.Typer(help="Welcome to SPAI")
app.add_typer(_auth.app, name="auth")
app.add_typer(list.app, name="list")


@app.command(help="Create a new project from starter template")
def init(
    project_name=typer.Argument(None, help="Project name"),
    path: Path = typer.Option(Path.cwd(), "-p", "--path", help="Project path"),
    template: str = typer.Option(None, "-t", "--template", help="Template name"),
    force: bool = typer.Option(False, "-f", "--force", help="Force download template"),
):
    try:
        if template:
            user = auth()
            path = download_template(user, template, path, force)
            return typer.echo(f"Project created at {path}")
        assert project_name is not None, "Project name is required."
        init_project(path, project_name)
        typer.echo(f"Project {project_name} created at {path}")
    except Exception as e:
        typer.echo(f"Error: {e}")


@app.command(help="Install requirements for a project")
def install(
    path: Path = typer.Argument(Path.cwd(), help="Project path"),
    verbose: bool = typer.Option(False, "-v", "--verbose"),
):
    try:
        install_requirements(path, typer, verbose)
    except Exception as e:
        typer.echo(f"Error: {e}")


@app.command(help="Run a project locally")
def run(
    path: Path = typer.Argument(Path.cwd(), help="Project path"),
    template: str = typer.Option(None, "-t", "--template", help="Template name"),
    template_path: Path = typer.Option(
        Path.cwd(),
        "-p",
        "--path",
        help="Destination path for the project created from the template",
    ),
    force: bool = typer.Option(False, "-f", "--force", help="Force download template"),
    install_reqs: bool = typer.Option(
        False, "-i", "--install-reqs", help="Install requirements"
    ),
    vars: List[str] = typer.Option(
        [], "--var", "-v", help="Variables to pass to the template"
    ),
):
    try:
        variables = parse_vars(vars)
        if template is not None:
            user = auth()
            path = download_template(user, template, template_path, force)
        if install_reqs:
            install_requirements(path, typer, False)
        return run_local(path, variables, typer)
    except Exception as e:
        typer.echo(f"Error: {e}")


@app.command(help="Deploy a project to the cloud")
def deploy(
    path: Path = typer.Argument(Path.cwd(), help="Project path"),
    template: str = typer.Option(None, "-t", "--template", help="Template name"),
    verbose: bool = typer.Option(False, "--verbose"),
    vars: List[str] = typer.Option(
        [], "--var", "-v", help="Variables to pass to the template"
    ),
):
    try:
        user = auth()
        variables = parse_vars(vars)
        if template:
            return deploy_template(user, template, variables, typer)
        return deploy_folder(user, path, variables, typer, verbose)
    except Exception as e:
        typer.echo(f"Error: {e}")


@app.command(help="Retrieve the logs of a service")
def logs(
    project: str = typer.Argument(..., help="Project name"),
    service_type: str = typer.Argument(
        ..., help="Service type (script, api, ui, etc.)"
    ),
    service_name: str = typer.Argument(..., help="Service name"),
):
    try:
        user = auth()
        logs = get_logs(user, project, service_type, service_name)
        typer.echo(logs)
    except Exception as e:
        typer.echo(f"Error: {e}")


@app.command(help="Clone a template")
def clone(
    template: str = typer.Argument(..., help="Template name"),
    path: Path = typer.Option(
        Path.cwd(),
        "-p",
        "--path",
        help="Destination path for the project created from the template",
    ),
    force: bool = typer.Option(False, "-f", "--force", help="Force download template"),
):
    try:
        user = auth()
        path = download_template(user, template, path, force)
        typer.echo(f"Template available at {path}.")
    except Exception as e:
        typer.echo(f"Error: {e}")


@app.command(help="Stop a project")
def stop(
    project: str = typer.Argument(None, help="Project name"),
    service_type: str = typer.Argument(None, help="Service type"),
    service_name: str = typer.Argument(None, help="Service name"),
    all: bool = typer.Option(False, "-a", "--all", help="Stop all services"),
    delete: bool = typer.Option(
        False, "-d", "--delete", help="Delete project after stopping"
    ),
):
    try:
        user = auth()
        if project is None:
            if not all:
                raise Exception(
                    "Project name is required (or use --all to stop all projects)."
                )
            projects = get_projects(user)
        else:
            projects = [project]
        for project in projects:
            services = get_services(user, project)
            if len(services) == 0:
                typer.echo(f"No services running in project '{project}'.")
                if delete:
                    delete_project(user, project)
                    typer.echo(f"Deleted project '{project}'.")
                return
            if service_type and service_name:
                stop_service_by_name(user, project, service_type, service_name)
                if delete:
                    delete_project(user, project)
                    typer.echo(f"Deleted project '{project}'.")
                return
            typer.echo(f"Stopping all services in project '{project}'...")
            for service in services:
                stop_service(user, project, service["id"])
            typer.echo(f"Stopped all services in project '{project}'.")
            if delete:
                delete_ok = typer.confirm(
                    "Are you sure you want to delete the project?"
                )
                if delete_ok:
                    delete_project(user, project)
                    typer.echo(f"Deleted project '{project}'.")
        return
    except Exception as e:
        return typer.echo(e)


@app.command(help="Delete a project")
def delete(
    project: str = typer.Argument(..., help="Project name"),
):
    try:
        user = auth()
        if project is None:
            raise Exception("Project name is required.")
        delete_project(user, project)
        return typer.echo(f"Deleted project '{project}'.")
    except Exception as e:
        return typer.echo(e)


@app.command(help="Get SPAI version")
def version():
    typer.echo(f"SPAI Version: {__version__}")


@app.command(help="Get SPAI API url and status")
def api():
    repo = APIRepo()
    typer.echo(f"SPAI API URL: {repo.url}")
    typer.echo(repo.info())


if __name__ == "__main__":
    app()



from pydantic import BaseModel
from typing import Union, List
from pathlib import Path
from typing import Optional
from collections import defaultdict

from .StorageConfig import LocalStorageConfig, S3StorageConfig


class ScriptConfig(BaseModel):
    name: str
    run_on_start: bool = True
    command: Optional[str] = None
    run_every: Optional[int] = None  # seconds (in cloud minutes)
    storage: Optional[str] = None  # folder to bind in cloud
    type: str = "script"
    wait_for: Optional[List[str]] = []

    # make sure that at least run_on_start or run_every is set
    def __init__(self, **data):
        super().__init__(**data)
        if not self.run_on_start and not self.run_every:
            raise ValueError(
                f"Script {self.name} must have either run_on_start or run_every set."
            )


class NotebookConfig(BaseModel):
    name: str
    command: Union[str, None] = None
    storage: Union[str, None] = None  # folder to bind in cloud
    port: int = 8888
    host: str = "0.0.0.0"
    type: str = "notebook"


class APIConfig(BaseModel):
    name: str
    command: Union[str, None] = None
    port: int = 8000
    host: str = "0.0.0.0"
    storage: Union[str, None] = None  # folder to bind in cloud
    type: str = "api"


class UIConfig(BaseModel):
    name: str
    command: str  # steamlit, javascript, ...
    port: int = 3000
    host: str = "0.0.0.0"
    env: dict = {}  # can accept the name of another service as a url placeholder
    type: str = "ui"


class Config(BaseModel):
    dir: Path
    project: str
    scripts: List[ScriptConfig] = []
    notebooks: List[NotebookConfig] = []
    apis: List[APIConfig] = []
    uis: List[UIConfig] = []
    storage: List[Union[LocalStorageConfig, S3StorageConfig]] = []

    def __init__(self, **data):
        super().__init__(**data)
        self.scripts = self.sort_scripts(self.scripts)

    # iterator for all the services
    def __iter__(self):
        # if self.storage:
        #     for storage in self.storage:
        #         yield storage
        if self.scripts:
            for script in self.scripts:
                yield script
        if self.notebooks:
            for notebook in self.notebooks:
                yield notebook
        if self.apis:
            for api in self.apis:
                yield api
        if self.uis:
            for ui in self.uis:
                yield ui
        if self.storage:
            for storage in self.storage:
                yield storage

    def type2folder(self, type):
        return type + "s"

    @staticmethod
    def sort_scripts(scripts):
        # cehcks
        object_names = {obj.name for obj in scripts}
        for obj in scripts:
            # check if all dependencies are in the scripts
            if not set(obj.wait_for).issubset(object_names):
                raise ValueError(
                    f"Dependency {set(obj.wait_for) - object_names} not found in scripts"
                )
            # check script is running on start
            if obj.wait_for and not obj.run_on_start:
                raise ValueError(
                    f"Script {obj.name} is not running on start, but has dependencies. This is not allowed."
                )
            # check all dependencies are running on start
            if obj.wait_for and not all(
                script.run_on_start for script in scripts if script.name in obj.wait_for
            ):
                raise ValueError(
                    f"Script {obj.name} has dependencies, but not all dependencies are running on start. This is not allowed."
                )
        # sort the scripts
        dependencies = {script.name: set(script.wait_for) for script in scripts}
        independent = {
            script.name for script in scripts if not dependencies[script.name]
        }
        sorted_scripts = []
        while independent:
            script = independent.pop()
            sorted_scripts.append(script)
            for name, deps in dependencies.items():
                deps.discard(script)
                if not deps and name not in sorted_scripts:
                    independent.add(name)
        if len(sorted_scripts) != len(scripts):
            raise ValueError("Dependency cycle detected in your scripts configuration.")
        return [
            next(script for script in scripts if script.name == name)
            for name in sorted_scripts
        ]



from .AWSDownloader import AWSDownloader

from typing import Optional


class AWSDEMDownloader(AWSDownloader):
    def __init__(self, aoi, datetime=None):
        super().__init__(aoi, datetime)

    def search_stac(self):
        items = self.catalog.search(
            bbox=self.aoi,
            collections=self.collection,
        ).item_collection()

        if len(items) == 0:
            print(f"No {self.collection} found for the AoI")
            return None

        return items

    def load_stac(
        self,
        chunks: Optional[dict] = {"time": 5, "x": 512, "y": 512},
    ):
        items = self.search_stac()
        if not items:
            return None
        data = self.load(
            items,
            chunks=chunks,
            crs=self.crs,
            resolution=self.resolution,
            bbox=self.aoi,
        )

        return data
